/*
 * arealight.sl - light source shader appropriate for an area light source.
 *
 * Description:
 *   Makes light with cosine falloff from the normal of the light source
 *   surface.  This is perfect for using with area light sources.
 *
 * Parameters:
 *   intensity, lightcolor - same meanings as pointlight
 *   shadowname - the name of the shadow map to use, if any
 *   shadowsamples - how many samples for shadow map sampling (larger is more
 *        expensive but higher quality)
 *   shadowblur - extra additive blur for shadow edges
 *   shadowbias - extra lookup bias to eliminate precision artifacts
 *        Bias < 0 means use the global value set by Option "shadow" "bias"
 *   
 * Hints:
 *   To use this light to ray trace, just pass shadowname="shadows".
 *
 * Entropy and BMRT are:
 * (c) Copyright 2001 Exluna, Inc.  Portions (c) 1990-2000 Larry Gritz.
 * All rights reserved.
 */

light
arealight (float  intensity = 1;
	   color  lightcolor = 1;
	   string shadowname = "";
	   float  shadowsamples = 1;
	   float  shadowblur = 0;
	   float  shadowbias = -1; )
{
#ifdef EXLUNA
    illuminate (P, N, 1.5707963 /* PI/2 */) {
	Cl = (intensity / (L.L)) * lightcolor;
#else
    /* This is so if we use one of these lights in a compatible
     * renderer that doesn't support area lights, it will
     * approximately work for a flat light source whose normal points
     * toward local +z.  
     */
    vector Nl = normalize (vector "shader" (0,0,1));
    illuminate (point "shader" (0,0,0), Nl, PI/2) {
	Cl = (intensity  * (Nl.normalize(L) / (L.L)) * lightcolor);
#endif
        if (shadowname != "")
            Cl *= 1 - shadow (shadowname, Ps, "samples", shadowsamples,
			      "blur", shadowblur, "bias", shadowbias);
    }
}
