/* fakecausticspotlight.sl - fake shadow-map-based caustics.
 *
 * Description:
 *  
 *   This shader is much light spotlight, except that instead of
 *   blocking light where the shadow map indicates occlusion, this light
 *   shines ONLY where the shadow map indicates occlusion.  By
 *   thresholding, blurring, and adding noise, this can produce a
 *   surprisingly accurate simulation of a refractive caustic.
 *
 * Parameters:
 *   intensity - basic scale on the intensity of the light
 *   lightcolor - color of the light emitted
 *   from, to - specify the direction of the light cone
 *   coneangle - half-width (in radians) of the illuminated cone
 *   conedeltaangle - controls the width (in radians) of the falloff
 *        as the angle approaches coneangle.
 *   beamdistribution - controls the angular falloff (from angular factors,
 *        independent of coneangle/conedeltaangle)
 *   shadowname - the name of the shadow map to use, if any
 *   shadowsamples - how many samples for shadow map sampling (larger is more
 *        expensive but higher quality)
 *   shadowblur - extra additive blur for shadow edges
 *   shadowbias - extra lookup bias to eliminate precision artifacts
 *        Bias < 0 means use the global value set by Option "shadow" "bias"
 *   threshold - controls thresholding for where the caustic appears.
 *        Larger numbers restrict the caustic toward the center of the
 *        shadow region.
 *   noiseamp, noisefreq, noisepow - control the noise of the caustic.
 *   
 * Entropy / BMRT are:
 * (c) Copyright 2001 Exluna, Inc.  All rights reserved.  
 */

light
fakecausticspotlight ( float  intensity = 1;
		       color  lightcolor = 1;
		       point  from = point "shader" (0,0,0);
		       point  to = point "shader" (0,0,1);
		       float  coneangle = radians(30);
		       float  conedeltaangle = radians(5);
		       float  beamdistribution = 2;
		       string shadowname = "";
		       float  shadowsamples = 1;
		       float  shadowblur = 0.01;
		       float  shadowbias = -1;
		       float  threshold = 0.5;
		       float  noiseamp = 0, noisefreq = 1, noisepow = 1;
    )
{
    uniform vector axis = normalize(to-from);

    illuminate (from, axis, coneangle) {
	float cosangle = (L . axis) / length(L);
	float atten = pow (cosangle, beamdistribution) / (L . L);
	atten *= smoothstep (cos(coneangle), cos(coneangle-conedeltaangle),
			     cosangle);
	Cl = atten * intensity * lightcolor;
        if (shadowname != "") {
            float caustic = shadow (shadowname, Ps, "samples", shadowsamples,
				    "blur", shadowblur, "bias", shadowbias);
	    caustic = smoothstep (threshold, 1, caustic);
	    if (noiseamp != 0) {
		point PL = transform ("shader", Ps);
		caustic *= noiseamp * pow (noise(PL*noisefreq), noisepow);
	    }
            Cl *= caustic;
	}
    }
}
