/****************************************************************************
 * fakeglass.sl -- Shiny reflective glass.
 *
 * Description:
 *   Makes semi-transparent glass, with environment reflections, but
 *   just using a fake opacity trick for transparency.  This is great
 *   for window glass, or any glass that is a thin enough sheet that you
 *   don't really need to see the warping of the refracted image.  It's
 *   much faster than ray tracing the refractions.
 * 
 * Parameters:
 *    Ka, Kd, Ks, roughness, specularcolor - The usual meaning
 *    Kr - coefficient for mirror-like reflections of environment
 *    blur - how blurry are the reflections? (0 = perfectly sharp)
 *    envname, envspace, envrad - controls for reflection
 *    Kt - coefficient for refracted transmission
 *    eta - the coefficient of refraction of the glass
 *
 * Hints:
 *    - With Entropy, you can ray trace the reflections by passing
 *      the parameters envname="reflection", envspace="current".
 *
 ***************************************************************************
 *
 * Entropy and BMRT are:
 * (c) Copyright 2001 Exluna, Inc.  Portions (c) 1990-2000 Larry Gritz.
 * All rights reserved.
 *
 * $Revision: 1.1 $    $Date: 2001/09/01 00:19:42 $
 *
 ****************************************************************************/


#include "material.h"


surface
fakeglass ( float Ka = 0.2, Kd = 0, Ks = 0.5, roughness = 0.05;
	    color specularcolor = 1;
	    float Kr = 1, reflblur = 0;
	    DECLARE_DEFAULTED_ENVPARAMS;
	    float Kt = 1, eta = 1.5; )
{
    normal Nf = faceforward (normalize(N), I);
    vector IN = normalize (I);

    /* Compute the reflection & refraction directions and amounts */
    vector Rfldir, Rfrdir;   /* Smooth reflection/refraction directions */
    float kr, kt;
    fresnel (IN, Nf, 1/eta, kr, kt, Rfldir, Rfrdir);
    kt = 1-kr;   /* Physically incorrect, but portable */
    kr *= Kr;
    kt *= Kt;
    /* Speedup -- at deep ray levels, reflection is unimportant */
    if (raylevel() > 0)
	kr = 0;
    /* Speedup for scanline -- don't shade back sides, but you HAVE to be sure
     * that normals correctly face outward.
     */
    if (N.I > 0 && raylevel() == 0)
	kr = kt = 0;

    /* Calculate the reflection & refraction color */
    color Crefl = Environment (envname, normalize(Rfldir), envspace, envrad,
			       reflblur, samples, kr);
    Ci = (Cs * (Ka*ambient() + Kd*diffuse(Nf)) +
	    (Crefl + Ks*LocIllumGlossy(Nf,-IN,roughness,0.5)));
    Oi = 1-kt;
}
